=head1 QualifierEditor::HitInfo

QualifierEditor::HitInfo - This object contains the information about one
search hit in the qualifier_editor application

$Header: /nfs/pathsoft/repository/psu/genlib/perl/src/QualifierEditor/HitInfo.pm,v 1.17 2001/11/05 14:10:56 kmr Exp $

=cut

package QualifierEditor::HitInfo;

use Exporter;
use strict;

use vars qw(@ISA @EXPORT_OK);

@ISA = qw(Exporter);

@EXPORT_OK = qw (get_protein_databases);

# change these variables to list the databases to search for the IDs
my $PROTEIN_DATABASES = "swall";

$PROTEIN_DATABASES =~ s/ /%20/g;

sub get_protein_databases
{
  return $PROTEIN_DATABASES;
}

my @srs_prot_libs = get_protein_databases ();
my $srs_prot_libs = "@srs_prot_libs";
my @srs_prot_libs_shortnames = qw"sw swnew tr trnew";
my $srs_prot_libs_shortnames = "@srs_prot_libs_shortnames";

sub new
{
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my %args = @_;

  my $self = { %args };

  $self->{cache} = {};

  bless $self, $class;

  return $self;
}

# return the organism name of the given $id using SRS
sub get_organism ($)
{
  my ($self) = @_;

  if (exists $self->{cache}{organism}) {
    return $self->{cache}{organism};
  }

  my $organism_lines = $self->get_field_from_id ("organism");

  if (!defined $organism_lines || $organism_lines =~ /^\s*$/) {

    if ($self->{id} =~ /^Y[A-P][LR]\d\d\d[WC]$/) {
      # it looks like a Yeast Y-code
      $self->{cache}{organism} = "Saccharomyces cerevisiae";
      return $self->{cache}{organism};
    }

    warn "can't find organism for ", $self->{id}, "\n";
    return undef;
  }

  if ($organism_lines =~ /OS\s+(.+)/) {
    my $organism = $1;

    $organism =~ s/\s*\(.*\s*\.\s*$//;

    $organism =~ s/\.\s*//;

    $self->{cache}{organism} = $organism;

    return $organism;

  } else {
    warn "can't find organism for ", $self->{id}, " got ", $organism_lines,
         " instead\n";
    return undef;
  }
}

# set $self->{cache}{des} to the DE lines for this id
sub get_des_lines
{
  my ($self) = @_;

  my $de_lines = $self->get_field_from_id ("des");

  $de_lines =~ s/^DE\s+(.*)/$1 /mg;

  $de_lines =~ s/\n/ /mg;
  $de_lines =~ s/\s+/ /g;
  $de_lines =~ s/\.\s*$//g;

  return $de_lines;
}


# return the product of the protein with the given $id using SRS
sub get_product
{
  my ($self) = @_;

  my $des = lc $self->get_des_lines ();

  my $return_value;

  my @things_to_uc =
    qw(n- o- d- l- dna dnaj atp gtp abc rna 60s 28s 18s 5s cgi cg atp gtp ctp
       ttp nadph nadp nadh nad);

  my %things_to_uc = (
                      atpase => "ATPase",
                      rrna => "rRNA",
                      cdna => "cDNA",
                     );

  if ($des =~ m/^(.*?)\s*\(/) {
    my $product = $1;

    $return_value = $product;
  } else {
    $return_value = $des;
  }

  my $gene_name = $self->get_gene_name;

  if (defined $gene_name) {
    my $capitalized_gene_name = $gene_name;

    if ($capitalized_gene_name =~ s/^(.)(.*)(.)$/\u$1\E\L$2\E\u$3\E/) {
      $return_value =~ s/$gene_name/$capitalized_gene_name/i;
    }
  }

  for my $word (@things_to_uc) {
    $return_value =~ s/\b$word\b/\U$word\E/;
  }

  for my $word (keys %things_to_uc) {
    my $new_word = $things_to_uc{$word};

    $return_value =~ s/\b$word\b/$new_word/i;
  }

  $return_value =~ s/\bkd\b/kD/;
  $return_value =~ s/\bkda\b/kDa/;

  return $return_value;
}

# return the EC number of the protein with the given $id using SRS
sub get_ecnumber
{
  my ($self) = @_;

  my $des = $self->get_des_lines ();

  if ($des =~ /\(EC\s*([^\)]+)\)/) {
    return $1;
  } else {
    return undef;
  }
}

# return the accession number of the protein with the given $id using SRS
sub get_acc_number
{
  my ($self) = @_;

  my $acc_line = $self->get_field_from_id ("acc");

  if (defined $acc_line && $acc_line =~ /^AC\s+([^ ;]+)/) {
    return $1;
  } else {
    return undef;
  }
}

# return the accession number of the EMBL entry that this protein came from
sub get_embl_acc_number
{
  my ($self) = @_;

  if (exists $self->{cache}{embl_acc}) {
    return $self->{cache}{embl_acc};
  }

  my $id = $self->{id};

  my $dbxref_line = $self->get_field_from_id ("dbxref");

  if ($dbxref_line =~ /^DR   EMBL; ([^;]+)/) {
    my $embl_acc = $1;

    $self->{cache}{embl_acc} = $embl_acc;

    return $embl_acc;
  }

  warn "can't find EMBL entry for $id\n";

  return "UNKNOWN ACCESSION";
}

# return the gene name of the protein with the given $id using SRS
sub get_gene_name ($)
{
  my ($self) = @_;

  my $gn_line = $self->get_field_from_id ("gn");

  if (defined $gn_line && $gn_line =~ /^gn\s+(.*?)\.?\s*$/i) {
    my $gene_name = $1;

    $gene_name =~ s/(.*)(.)/\L$1\E\U$2/;

    return $gene_name;
  } else {
    return undef;
  }
}

# return a data field for this ID from SRS
sub get_field_from_id
{
  my ($self, $field_name) = @_;

  if (exists $self->{cache}{$field_name}) {
    return $self->{cache}{$field_name};
  }

  my $id = $self->{id};

  my $getz_command =
    "getz -f $field_name \"[libs={$srs_prot_libs}-id: $id] | [libs-AccNumber: $id ] | [libs-ProteinID: $id] \" |";

  # special case for tremblnew entries - check tremblnew and if the entry
  # isn't there, get the entry from TREMBL via the dbxref
  if ($id =~ /[a-z][a-z][a-z]\d\d\d\d\d/i) {

    system "efetch trnew:$id > /dev/null 2> /dev/null";

    if ($?) {
      # it failed so try using dbxref

      $getz_command =
        "getz -f $field_name \"[\{$srs_prot_libs\}-dbxref:$id*]\" |";
    }
  }

  open (GETZ, $getz_command) or die "can't open pipe to getz\n";
  local $/ = undef;
  my $field = <GETZ>;
  close GETZ or die "can't close pipe to getz\n";

  chomp $field;

  $self->{cache}{$field_name} = $field;

  return $field;
}

1;
